/////////////////////////////////////////////////////////////////////////////
// $Id: Matrice.cpp,v 1.3 1997/08/20 02:56:35 parizeau Exp $
//
//   Auteurs:      Marc Parizeau
//   Date:         6/3/96
//   Modification: 14 mars 1996 ---> ajout de Moy(),Var(),CoVar(),Correl()
//   Modification: 19 mars 1996 ---> ajout de "const" pour plusieurs operators
//   Modification: 20 mars 1996 ---> ajout de Vecteur(Matrice&)
//   Modification: 13 mai  1996 ---> ajout de Vecteur(Table1D&)
//
//   Description:  Methodes de la classe Matrice.
/////////////////////////////////////////////////////////////////////////////

#include <memory.h>
#include "Matrice.hpp"

/////////////////////////////////////////////////////////////////////////////
//  Matrice: Construire Matrice ixi avec valeur initiale v.
/////////////////////////////////////////////////////////////////////////////
Matrice::Matrice(int _i,int _j,double _v) : Table2D<double>(_i,_j) {
  Initialiser(_v);
}

/////////////////////////////////////////////////////////////////////////////
//  Matrice: Construire Matrice nxn identite.
/////////////////////////////////////////////////////////////////////////////
Matrice::Matrice(Type _t,int _n) : Table2D<double>(_n,_n) {
  if(_t == IDENTITE) Identite();
}

///////////////////////////////////////////////////////////////////////////
// operator~: Transposer la Table2D.
///////////////////////////////////////////////////////////////////////////
Matrice Matrice::operator~() const { 
  Matrice res(Cols(),Lignes());
  for(int i=0;i<Lignes();i++) {
    for(int j=0;j<Cols();j++) res(j,i) = (*this)(i,j);
  }
  return res;
}

///////////////////////////////////////////////////////////////////////////
// operator+: Additionner avec une Matrice.
///////////////////////////////////////////////////////////////////////////
Matrice Matrice::operator+(const Matrice &_mat) const {
  if(Lignes() != _mat.Lignes() || Cols() != _mat.Cols())
     throw ErrBase("Matrice incompatible!");
  Matrice res(Lignes(),Cols());
  for(int i=0;i<Lignes();i++) {
	for(int j=0;j<Cols();j++) res(i,j) = (*this)(i,j) + _mat(i,j);
  }	
  return res;
}

///////////////////////////////////////////////////////////////////////////
// operator+=: Additionner/Affecter avec une Matrice.
///////////////////////////////////////////////////////////////////////////
Matrice &Matrice::operator+=(const Matrice &_mat)  {
  if(Lignes() != _mat.Lignes() || Cols() != _mat.Cols())
     throw ErrBase("Matrice incompatible!");
  for(int i=0;i<Lignes();i++) {
    for(int j=0;j<Cols();j++) (*this)(i,j) += _mat(i,j);
  }
  return *this;
}

///////////////////////////////////////////////////////////////////////////
// operator+: Additionner avec un scalaire.
///////////////////////////////////////////////////////////////////////////
Matrice Matrice::operator+(double _scal) const {
  Matrice res(Lignes(),Cols());
  for(int i=0;i<Lignes();i++) {
    for(int j=0;j<Cols();j++) res(i,j) = (*this)(i,j) + _scal;
  }
  return res;
}

///////////////////////////////////////////////////////////////////////////
// operator+=: Additionner/Affecter avec un scalaire.
///////////////////////////////////////////////////////////////////////////
Matrice &Matrice::operator+=(double _scal)  {
  for(int i=0;i<Lignes();i++) {
    for(int j=0;j<Cols();j++) (*this)(i,j) += _scal;
  }
  return *this;
}

///////////////////////////////////////////////////////////////////////////
// operator-: Soustraire avec une Matrice.
///////////////////////////////////////////////////////////////////////////
Matrice Matrice::operator-(const Matrice &_mat) const {
  if(Lignes() != _mat.Lignes() || Cols() != _mat.Cols())
     throw ErrBase("Matrice incompatible!");
  Matrice res(Lignes(),Cols());
  for(int i=0;i<Lignes();i++) {
    for(int j=0;j<Cols();j++) res(i,j) = (*this)(i,j) - _mat(i,j);
  }
  return res;
}

///////////////////////////////////////////////////////////////////////////
// operator-=: Soustraire/Affecter avec une Matrice.
///////////////////////////////////////////////////////////////////////////
Matrice &Matrice::operator-=(const Matrice &_mat)  {
  if(Lignes() != _mat.Lignes() || Cols() != _mat.Cols())
     throw ErrBase("Matrice incompatible");
  for(int i=0;i<Lignes();i++) {
    for(int j=0;j<Cols();j++) (*this)(i,j) -= _mat(i,j);
  }
  return *this;
}

///////////////////////////////////////////////////////////////////////////
// operator-: Soustraire avec un scalaire.
///////////////////////////////////////////////////////////////////////////
Matrice Matrice::operator-(double _scal) const {
  Matrice res(Lignes(),Cols());
  for(int i=0;i<Lignes();i++) {
    for(int j=0;j<Cols();j++) res(i,j) = (*this)(i,j) - _scal;
  }
  return res;
}

///////////////////////////////////////////////////////////////////////////
// operator-=: Soustraire/Affecter avec un scalaire.
///////////////////////////////////////////////////////////////////////////
Matrice &Matrice::operator-=(double _scal)  {
  for(int i=0;i<Lignes();i++) {
    for(int j=0;j<Cols();j++) (*this)(i,j) -= _scal;
  }
  return *this;
}

///////////////////////////////////////////////////////////////////////////
// operator*: Multiplier avec une Matrice.
///////////////////////////////////////////////////////////////////////////
Matrice Matrice::operator*(const Matrice &_mat) const {
  if(Cols() != _mat.Lignes())
     throw ErrBase("Matrice incompatible!");
  Matrice res(Lignes(),_mat.Cols());
  for(int i=0;i<res.Lignes();i++) {
    for(int j=0;j<res.Cols();j++) {
      for(int k=0;k<Cols();k++) res(i,j) += (*this)(i,k) * _mat(k,j);
    }
  }
  return res;
}

///////////////////////////////////////////////////////////////////////////
// operator*: Multiplier avec un scalaire.
///////////////////////////////////////////////////////////////////////////
Matrice Matrice::operator*(double _scal) const {
  Matrice res(Lignes(),Cols());
  for(int i=0;i<Lignes();i++) {
    for(int j=0;j<Cols();j++) res(i,j) = (*this)(i,j) * _scal;
  }
  return res;
}

///////////////////////////////////////////////////////////////////////////
// operator*=: Multiplier/Affecter avec un scalaire.
///////////////////////////////////////////////////////////////////////////
Matrice &Matrice::operator*=(double _scal)  {
  for(int i=0;i<Lignes();i++) {
    for(int j=0;j<Cols();j++) (*this)(i,j) *= _scal;
  }
  return *this;
}

///////////////////////////////////////////////////////////////////////////
// operator/: Diviser par un scalaire.
///////////////////////////////////////////////////////////////////////////
Matrice Matrice::operator/(double _scal) const {
  Matrice res(Lignes(),Cols());
  for(int i=0;i<Lignes();i++) {
    for(int j=0;j<Cols();j++) res(i,j) = (*this)(i,j) / _scal;
  }
  return res;
}

///////////////////////////////////////////////////////////////////////////
// operator/=: Diviser/Affecter par un scalaire.
///////////////////////////////////////////////////////////////////////////
Matrice &Matrice::operator/=(double _scal)  {
  for(int i=0;i<Lignes();i++) {
    for(int j=0;j<Cols();j++) (*this)(i,j) /= _scal;
  }
  return *this;
}

////////////////////////////////////////////////////////////////////////////
// ArgMax: Retourner les indices de la valeur maximum pour la matrice.
/////////////////////////////////////////////////////////////////////////////
void Matrice::ArgMax(int &_i,int &_j) const {
  _i = 0,_j=0;
  for(int i=0;i<Lignes();i++) {
    for(int j=0;j<Cols();j++) {
      if((*this)(i,j) > (*this)(_i,_j)) {
	_i = i;
	_j = j;
      }
    }
  }
}

////////////////////////////////////////////////////////////////////////////
// ArgMin: Retourner les indices de la valeur minimum pour la matrice.
/////////////////////////////////////////////////////////////////////////////
void Matrice::ArgMin(int &_i,int &_j) const {
  _i = 0,_j=0;
  for(int i=0;i<Lignes();i++) {
    for(int j=0;j<Cols();j++) {
      if((*this)(i,j) < (*this)(_i,_j)) {
	_i = i;
	_j = j;
      }
    }
  }
}

////////////////////////////////////////////////////////////////////////////
// ArgMaxCol: Retourner l'indice de la valeur maximum pour la colonne.
/////////////////////////////////////////////////////////////////////////////
int Matrice::ArgMaxCol(int j) const {
  int _i = 0;
  for(int i=1;i<Lignes();i++) if((*this)(i,j) > (*this)(_i,j)) _i = i;
  return _i;
}

////////////////////////////////////////////////////////////////////////////
// ArgMinCol: Retourner l'indice de la valeur minimum pour la colonne.
/////////////////////////////////////////////////////////////////////////////
int Matrice::ArgMinCol(int j) const {
  int _i = 0;
  for(int i=1;i<Lignes();i++) if((*this)(i,j) < (*this)(_i,j)) _i = i;
  return _i;
}

////////////////////////////////////////////////////////////////////////////
// ArgMaxLigne: Retourner l'indice de la valeur maximum pour la ligne.
/////////////////////////////////////////////////////////////////////////////
int Matrice::ArgMaxLigne(int i) const {
  int _j = 0;
  for(int j=1;j<Cols();j++) if((*this)(i,j) > (*this)(i,_j)) _j = j;
  return _j;
}

////////////////////////////////////////////////////////////////////////////
// ArgMinLigne: Retourner l'indice de la valeur minimum pour la ligne.
/////////////////////////////////////////////////////////////////////////////
int Matrice::ArgMinLigne(int i) const {
  int _j = 0;
  for(int j=1;j<Cols();j++) if((*this)(i,j) < (*this)(i,_j)) _j = j;
  return _j;
}

/////////////////////////////////////////////////////////////////////////////
// Correl: Calculer la matrice de correlation.
/////////////////////////////////////////////////////////////////////////////
Matrice Matrice::Correl() const {
  Matrice mcr(*this);
  Vecteur moy(Cols());
  Vecteur var(Cols());
  int i,j;
  for(j=0;j<Cols();j++) {
    moy(j) = MoyCol(j);
    var(j) = VarCol(j);
  }
  for(i=0;i<Lignes();i++) {
    for(j=0;j<Cols();j++) mcr(i,j) = ((*this)(i,j)-moy(j))/sqrt(var(j));
  }
  return (~mcr*mcr)/(Lignes()-1);
}

/////////////////////////////////////////////////////////////////////////////
// CoVar: Calculer la matrice de covariance.
/////////////////////////////////////////////////////////////////////////////
Matrice Matrice::CoVar() const {
  Matrice mc(*this);
  Vecteur moy(Cols());
  int i,j;
  for(j=0;j<Cols();j++) moy(j) = MoyCol(j);
  for(i=0;i<Lignes();i++) {
    for(j=0;j<Cols();j++) mc(i,j) = (*this)(i,j)-moy(j);
  }
  return (~mc*mc)/(Lignes()-1);
}

/////////////////////////////////////////////////////////////////////////////
// Det: Calculer le determinant de la matrice. (L'algorithme provient
//      du livre "Numerical Recipes in C", section 2.3)
/////////////////////////////////////////////////////////////////////////////
double Matrice::Det() const {
  if(Lignes() != Cols()) throw ErrBase("Matrice incompatible!");
  Table1D<int> indx(Lignes());
  int d;
  Matrice decompose = LUDec(indx,d);
  double determinant = d;
  for(int i=0;i<Lignes();i++) determinant *= decompose(i,i);
  return determinant;
}

/////////////////////////////////////////////////////////////////////////////
// Identite: Initialiser a la matrice identite.
/////////////////////////////////////////////////////////////////////////////
void Matrice::Identite() {
  if(Lignes() != Cols()) throw ErrBase("Matrice incompatible!");
  Initialiser(0);
  for(int i=0;i<Lignes();i++) (*this)(i,i) = 1;
}

/////////////////////////////////////////////////////////////////////////////
// Inv: Retourner la matrice inverse. (L'algorithme provient
//      du livre "Numerical Recipes in C", section 2.3)
/////////////////////////////////////////////////////////////////////////////
Matrice Matrice::Inv() const {
  if(Lignes() != Cols()) throw ErrBase("Matrice incompatible!");
  Matrice Y(IDENTITE,Lignes());
  Table1D<int> indx(Lignes());
  Vecteur B(Lignes());
  int d;
  Matrice decompose = LUDec(indx,d);
  for(int j=0;j<Cols();j++) {
    B.CopieCol(Y,j,0);
    decompose.LUBSub(indx,B);
    Y.CopieCol(B,0,j);
  }
  return Y;
}

////////////////////////////////////////////////////////////////////////////
// Max: Retourner la valeur maximum de la matrice.
/////////////////////////////////////////////////////////////////////////////
double Matrice::Max() const {
  double max = (*this)(0,0),tmp;
  for(int i=0;i<Lignes();i++) {
    for(int j=0;j<Cols();j++) if((tmp=(*this)(i,j)) > max) max = tmp;
  }
  return max;
}

////////////////////////////////////////////////////////////////////////////
// MaxCol: Retourner la valeur maximum dans la colonne specifiee.
/////////////////////////////////////////////////////////////////////////////
double Matrice::MaxCol(int _j) const {
  double max = (*this)(0,_j),tmp;
  for(int i=1;i<Lignes();i++) if((tmp=(*this)(i,_j)) > max) max = tmp;
  return max;
}

////////////////////////////////////////////////////////////////////////////
// MaxLigne: Retourner la valeur maximum dans la ligne specifiee.
/////////////////////////////////////////////////////////////////////////////
double Matrice::MaxLigne(int _i) const {
  double max = (*this)(_i,0),tmp;
  for(int j=1;j<Cols();j++) if((tmp=(*this)(_i,j)) > max) max = tmp;
  return max;
}

////////////////////////////////////////////////////////////////////////////
// Min: Retourner la valeur minimum de la matrice.
/////////////////////////////////////////////////////////////////////////////
double Matrice::Min() const {
  double min = (*this)(0,0),tmp;
  for(int i=0;i<Lignes();i++) {
    for(int j=0;j<Cols();j++) if((tmp=(*this)(i,j)) < min) min = tmp;
  }
  return min;
}

////////////////////////////////////////////////////////////////////////////
// MinCol: Retourner la valeur minimum dans la colonne specifiee.
/////////////////////////////////////////////////////////////////////////////
double Matrice::MinCol(int _j) const {
  double min = (*this)(0,_j),tmp;
  for(int i=1;i<Lignes();i++) if((tmp=(*this)(i,_j)) < min) min = tmp;
  return min;
}

////////////////////////////////////////////////////////////////////////////
// MinLigne: Retourner la valeur minimum dans la ligne specifiee.
/////////////////////////////////////////////////////////////////////////////
double Matrice::MinLigne(int _i) const {
  double min = (*this)(_i,0),tmp;
  for(int j=1;j<Cols();j++) if((tmp=(*this)(_i,j)) < min) min = tmp;
  return min;
}

/////////////////////////////////////////////////////////////////////////////
// Moy: Retourner la moyenne de tous les elements.
/////////////////////////////////////////////////////////////////////////////
double Matrice::Moy() const {return Somme()/Lignes()/Cols();}

/////////////////////////////////////////////////////////////////////////////
// MoyCol: Retourner la moyenne sur la colonne specifiee.
/////////////////////////////////////////////////////////////////////////////
double Matrice::MoyCol(int _j) const {return SommeCol(_j)/Lignes();}

/////////////////////////////////////////////////////////////////////////////
// MoyLigne: Retourner la moyenne sur la ligne specifiee.
/////////////////////////////////////////////////////////////////////////////
double Matrice::MoyLigne(int _i) const {return SommeLigne(_i)/Cols();}

////////////////////////////////////////////////////////////////////////////
// Somme: Faire la somme des elements de la matrice.
////////////////////////////////////////////////////////////////////////////
double Matrice::Somme() const {
   double res = 0;
   for (int j=0;j<Cols();j++) res += SommeCol(j);
   return res;
}

////////////////////////////////////////////////////////////////////////////
// SommeCol: Faire la somme des elements d'une colonne de la matrice.
////////////////////////////////////////////////////////////////////////////
double Matrice::SommeCol(int _j) const {
   double res = 0;
   for (int i=0;i<Lignes();i++) res += (*this)(i,_j);
   return res;
}

////////////////////////////////////////////////////////////////////////////
// SommeLigne: Faire la somme des elements d'une ligne de la matrice.
////////////////////////////////////////////////////////////////////////////
double Matrice::SommeLigne(int _i) const {
   double res = 0;
   for (int j=0;j<Cols();j++) res += (*this)(_i,j);
   return res;
}

////////////////////////////////////////////////////////////////////////////
// Somme2: Faire la somme du carre des elements de la matrice.
////////////////////////////////////////////////////////////////////////////
double Matrice::Somme2() const {
  double res = 0;
  for(int j=0;j<Cols();j++) res += Somme2Col(j);
  return res;
}

////////////////////////////////////////////////////////////////////////////
// Somme2Col: Faire somme du carre des elements d'une colonne de la matrice.
////////////////////////////////////////////////////////////////////////////
double Matrice::Somme2Col(int _j) const {
   double res = 0;
   for (int i=0;i<Lignes();i++) res += ((*this)(i,_j)) * ((*this)(i,_j));
   return res;
}

////////////////////////////////////////////////////////////////////////////
// Somme2Ligne: Faire somme du carre des elements d'une ligne de la matrice.
////////////////////////////////////////////////////////////////////////////
double Matrice::Somme2Ligne(int _i) const {
   double res = 0;
   for (int j=0;j<Cols();j++) res += ((*this)(_i,j)) * ((*this)(_i,j));
   return res;
}

////////////////////////////////////////////////////////////////////////////
// SubCols: Substituer les colonnes c1 et c2.
///////////////////////////////////////////////////////////////////////////
void Matrice::SubCols(int _c1,int _c2) {
  double tmp;
  for(int i=0;i<Lignes();i++) {
    tmp = (*this)(i,_c1);
    (*this)(i,_c1) = (*this)(i,_c2);
    (*this)(i,_c2) = tmp;
  }
}

////////////////////////////////////////////////////////////////////////////
// SubLignes: Substituer les lignes l1 et l2.
///////////////////////////////////////////////////////////////////////////
void Matrice::SubLignes(int _l1,int _l2) {
  double tmp;
  for(int j=0;j<Cols();j++) {
    tmp = (*this)(_l1,j);
    (*this)(_l1,j) = (*this)(_l2,j);
    (*this)(_l2,j) = tmp;
  }
}

/////////////////////////////////////////////////////////////////////////////
// Var: Retourner la variance de tous les elements.
/////////////////////////////////////////////////////////////////////////////
double Matrice::Var() const {
  double mu = Moy();
  return (Somme2() - (Lignes()*Cols()*mu*mu))/(Lignes()*Cols()-1);
}

/////////////////////////////////////////////////////////////////////////////
// VarCol: Retourner la variance sur la colonne specifiee.
/////////////////////////////////////////////////////////////////////////////
double Matrice::VarCol(int _j) const {
  double mu = MoyCol(_j);
  return (Somme2Col(_j) - (Lignes()*mu*mu))/(Lignes()-1);
}

/////////////////////////////////////////////////////////////////////////////
// VarLigne: Retourner la variance sur la ligne specifiee.
/////////////////////////////////////////////////////////////////////////////
double Matrice::VarLigne(int _i) const {
  double mu = MoyLigne(_i);
  return (Somme2Ligne(_i) - (Cols()*mu*mu))/(Cols()-1);
}

////////////////////////////////////////////////////////////////////////////
// CopieCol: Copier cols c1 de mat vers cols c2 de *this.
///////////////////////////////////////////////////////////////////////////
void Matrice::CopieCol(Matrice &_mat,int _c1,int _c2) {
  for(int i=0;i<Lignes();i++) (*this)(i,_c2) = _mat(i,_c1);
}

////////////////////////////////////////////////////////////////////////////
// Scale: Calculer le facteur d'echelle pour la decomposition L-U.
///////////////////////////////////////////////////////////////////////////
Table1D<double> Matrice::Scale() const {
  double tmp;
  Table1D<double> res(Cols());
  for(int i=0;i<Lignes();i++) {
    double maximum = 0;
    for(int j=0;j<Cols();j++) {
      if((tmp = fabs((*this)(i,j))) > maximum) maximum = tmp;
    }
    if(maximum == 0) throw ErrBase("Matrice singuliere!");
    res[i] = 1/maximum;
  }
  return res;
}

const double TINY = double(1e-20);

////////////////////////////////////////////////////////////////////////////
// LUDec: Decomposition L-U.
///////////////////////////////////////////////////////////////////////////
Matrice Matrice::LUDec(Table1D<int> &_indx,int &_d) const {
  _d = 1;
  int i,j,k,l;
  double dum,sum,maximum;
  Matrice decompose(*this);
  Table1D<double> scale = Scale();
  for(j=0;j<Cols();j++) {
    for(i=0;i<j;i++) {
      sum = decompose(i,j);
      for(k=0;k<i;k++) sum -= decompose(i,k) * decompose(k,j);
      decompose(i,j) = sum;
    }
    maximum = 0;
    for(i=j;i<Lignes();i++) {
      sum = decompose(i,j);
      for(k=0;k<j;k++) sum -= decompose(i,k) * decompose(k,j);
      decompose(i,j) = sum;
      if((dum = scale[i]*fabs(sum)) >= maximum) {
        l = i;
        maximum = dum;
      }
    }
    if(j != l) {
      decompose.SubLignes(l,j);
      _d = -_d;
      scale[l] = scale[j];
     }
    _indx[j] = l;
    if(decompose(j,j) == 0) decompose(j,j) = TINY;
    if(j != Cols()-1) {
      dum = 1.0/decompose(j,j);
      for(i=j+1;i<Lignes();i++) decompose(i,j) *= dum;
    }
  }
  return decompose;
}

////////////////////////////////////////////////////////////////////////////
// LUBSub: Back Substitution pour la decomposition L-U.
///////////////////////////////////////////////////////////////////////////
void Matrice::LUBSub(Table1D<int> &_indx,Matrice &_b) {
  int i,ii=-1,ip,j;
  double sum;
  for(i=0;i<Lignes();i++) {
    ip = _indx[i];
    sum = _b(ip,0);
    _b(ip,0) = _b(i,0);
    if(ii >= 0) for(j=ii;j<i;j++) sum -= (*this)(i,j) * _b(j,0);
    else if(sum) ii = i;
    _b(i,0) = sum;
  }
  for(i=Lignes()-1;i>=0;i--) {
    sum = _b(i,0);
    for(j=i+1;j<Cols();j++) sum -= (*this)(i,j) * _b(j,0);
    _b(i,0) = sum/(*this)(i,i);
  }
}

///////////////////////////////////////////////////////////////////////////
// Vecteur: Convertir une Matrice en Vecteur.
///////////////////////////////////////////////////////////////////////////
Vecteur::Vecteur(const Matrice &_mat)  {
  if(_mat.Cols() != 1) throw ErrBase("Matrice incompatible!");
  LignesCols(_mat.Lignes(),1);
  memcpy(&(*this)[0],&_mat[0],Dim()*sizeof(double));
}

///////////////////////////////////////////////////////////////////////////
// Vecteur: Convertir une Table1D<double> en Vecteur.
///////////////////////////////////////////////////////////////////////////
Vecteur::Vecteur(const Table1D<double> &_tb)  {
  LignesCols(_tb.Dim(),1);
  memcpy(&(*this)[0],&_tb[0],Dim()*sizeof(double));
}
