/////////////////////////////////////////////////////////////////////////////
// $Id: Matrice.hpp,v 1.2 1997/03/06 21:21:50 parizeau Exp $
//
//   Auteurs:      Marc Parizeau
//   Creation:     6/3/96
//
//   Description:  Classe permettant de gerer dynamiquement des
//                 matrices de nombres reels.
/////////////////////////////////////////////////////////////////////////////

#ifndef MATRICE_HPP_
#define MATRICE_HPP_

#include <iostream.h>
#include <iomanip.h>
#include <math.h>
#include "Table2D.hpp"

/////////////////////////////////////////////////////////////////////////////

class Matrice : public Table2D<double> {

///////////////////// Debut de l'interface publique /////////////////////////

 public:   enum Type {IDENTITE};

           Matrice() {}                     // Construire Matrice 0x0
           Matrice(int i,int j,double v=0); // Construire Matrice ixj
           Matrice(Type,int dim);           // Construire Matrice identite

  Matrice  operator~() const;               // Transposer
  Matrice  operator+(const Matrice &) const;// Additionner Matrice
  Matrice &operator+=(const Matrice &);     // Additionner/Affecter Matrice
  Matrice  operator+(double) const;         // Additionner scalaire
  Matrice &operator+=(double);              // Additionner/Affecter scalaire
  Matrice  operator-(const Matrice &) const;// Soustraire Matrice
  Matrice &operator-=(const Matrice &);     // Soustraire/Affecter Matrice
  Matrice  operator-(double) const;         // Soustraire scalaire
  Matrice &operator-=(double);              // Soustraire/Affecter scalaire
  Matrice  operator*(const Matrice &) const;// Multiplier Matrice
  Matrice  operator*(double) const;         // Multiplier scalaire
  Matrice &operator*=(double);              // Multiplier/Affecter scalaire
  Matrice  operator/(double) const;         // Diviser scalaire
  Matrice &operator/=(double);              // Diviser/Affecter scalaire

  void     ArgMax(int &i,int &j) const;     // Retourner indices max
  void     ArgMin(int &i,int &j) const;     // Retourner indices min
  int      ArgMaxCol(int=0) const;          // Retourner indice max colonne
  int      ArgMinCol(int=0) const;          // Retourner indice min colonne
  int      ArgMaxLigne(int) const;          // Retourner indice max ligne
  int      ArgMinLigne(int) const;          // Retourner indice min ligne
  void     CopieCol(Matrice &,int,int);     // Copier colonne de matrice
  Matrice  Correl() const;                  // Retourner matrice correlation
  Matrice  CoVar() const;                   // Retourner matrice covariance
  double   Det() const;                     // Calculer determinant
  void     Identite();                      // Initialiser identite
  Matrice  Inv() const;                     // Retourner inverse
  double   Max() const;                     // Retourner max
  double   MaxCol(int=0) const;             // Retourner max colonne
  double   MaxLigne(int) const;             // Retourner max ligne
  double   Min() const;                     // Retourner min
  double   MinCol(int=0) const;             // Retourner min colonne
  double   MinLigne(int) const;             // Retourner min ligne
  double   Moy() const;                     // Retourner moyenne
  double   MoyCol(int=0) const;             // Retourner moyenne colonne
  double   MoyLigne(int) const;             // Retourner moyenne ligne
  double   Somme() const;                   // Sommer elements
  double   SommeCol(int=0) const;           // Sommer elements colonne
  double   SommeLigne(int) const;           // Sommer elements ligne
  double   Somme2() const;                  // Sommer carre elements
  double   Somme2Col(int=0) const;          // Sommer carre elements colonne
  double   Somme2Ligne(int) const;          // Sommer carre elements ligne
  void     SubCols(int,int);                // Substituer deux colonnes
  void     SubLignes(int,int);              // Substituer deux lignes
  double   Var() const;                     // Retourner variance
  double   VarCol(int=0) const;             // Retourner variance colonne
  double   VarLigne(int) const;             // Retourner variance ligne

friend ostream &operator<<(ostream &_os,const Matrice &_mat) 
  {_mat.Ecrire(_os); return _os;}
friend istream &operator>>(istream &_is,Matrice &_mat) 
  {_mat.Lire(_is); return _is;}

///////////////////// Fin de l'interface publique //////////////////////////

 private:                                   // fonctions pour Inv et Det
  Matrice  LUDec(Table1D<int>&,int &) const;// Decomposer L-U
  void     LUBSub(Table1D<int>&,Matrice&);  // L-U back substitution
  Table1D<double> Scale() const;            // Calculer facteurs d'echelle

};

/////////////////////////////////////////////////////////////////////////////

class Vecteur : public Matrice {

///////////////////// Debut de l'interface publique /////////////////////////

public:
  Vecteur(int i=0):Matrice(i,1) {} // Construire Matrice ix1
  Vecteur(const Matrice&);         // Convertir Matrice->Vecteur
  Vecteur(const Table1D<double>&); // Convertir Table1D->Vecteur

  inline double &operator()(int i) const {return Matrice::operator()(i,0);}

  inline void Dim(int n) {LignesCols(n,1);}
  inline int  Dim() const {return Lignes();}

///////////////////// Fin de l'interface publique //////////////////////////

private:
  void     operator=(int);

};

#endif
