//////////////////////////////////////////////////////////////////////////////
//: $Id: Perceptron.hpp,v 1.10 1998/04/15 21:22:36 parizeau Exp $
//
//  Auteurs:        Marc Parizeau et Jean-Francois Hebert
//  Date:           08/01/97
//  Description:    Classe permettant de gerer un perceptron multi-couches. 
//  Voir: http://www.gel.ulaval.ca/~parizeau/C++/classes/doc/algo2/algo2.html
//////////////////////////////////////////////////////////////////////////////

#ifndef PERCEPTRON_H_
#define PERCEPTRON_H_

#include "Matrice.hpp"
#include "stdafx.h"
#include "resource.h"
#include "ProgDlg.h"

/////////////////////////////////////////////////////////////////////////////

class PoidsP {

///////////////////// Debut de l'interface publique /////////////////////////
    
public:
  PoidsP() {cour = prec = _gnpa.Unif()-0.5;};

  int operator==(const PoidsP &_p) {return (cour == _p.cour) ? 1 : 0;}
  int operator!=(const PoidsP &_p) {return (cour != _p.cour) ? 1 : 0;}
  int operator< (const PoidsP &_p) {return (cour <  _p.cour) ? 1 : 0;}
  int operator> (const PoidsP &_p) {return (cour >  _p.cour) ? 1 : 0;}

  double Poids() {return cour;}                 // Retourner valeur courante
  void   Poids(double _v) {prec=cour; cour=_v;} // Definir nouvelle valeur
  double VarPoids() {return cour-prec;}         // Retourner variation poids

///////////////////// Fin de l'interface publique //////////////////////////

private:
  double cour;                                  // poids courant
  double prec;                                  // poids precedent
  
  friend ostream &operator<<(ostream &_os,const PoidsP &_poids) {
    _os << _poids.cour; return _os;
  };
  friend istream &operator>>(istream &_is,PoidsP &_poids) {
    _is >> _poids.cour; _poids.prec = _poids.cour; return _is;
  };

};

class CoucheP;

/////////////////////////////////////////////////////////////////////////////

class NeuroneP : public Table1D<PoidsP> {

///////////////////// Debut de l'interface publique /////////////////////////
    
public:
  NeuroneP() {delta = 0; sortie = 0;}

  void   Activer(CoucheP &prec);                 // Activer sortie (n. cache)
  void   Activer(const Vecteur &in);             // Activer sortie (n. entree)
  double Biais() {return biais.Poids();}         // Retourner biais
  void   Biais(double _val) {biais.Poids(_val);} // Definir poids du biais
  double Delta() {return delta;}                 // Retourner le delta
  void   Delta(double _val) {delta = _val;}      // Definir le delta
  double Sigmoid(double);                        // Fonction d'activation
  double Sortie() {return sortie;}               // Retourner sortie
  double VarBiais() {return biais.VarPoids();}   // Retourner variation biais

///////////////////// Fin de l'interface publique //////////////////////////

private:
  PoidsP biais;                                   // biais
  double delta;                                   // erreur attribuee
  double sortie;                                  // valeur de sortie
  
  friend ostream &operator<<(ostream &_os,const NeuroneP &_neurone) {
    _os << _neurone.biais << endl; _neurone.Ecrire(_os); return _os;
  };
  friend istream &operator>>(istream &_is,NeuroneP &_neurone) {
    _is >> _neurone.biais; _neurone.Lire(_is); return _is;
  };
  
};

/////////////////////////////////////////////////////////////////////////////

class CoucheP : public Table1D<NeuroneP> {};

/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////

class Perceptron : public Table1D<CoucheP> {

///////////////////// Debut de l'interface publique /////////////////////////
    
public:
  Perceptron(int mgraphe=0);                        // Construire sans couches
  Perceptron(TableInt &couches,int mgraphe=0);      // Construire avec couches
  ~Perceptron();                                    // Detruire les fenetres

  struct Data {                                     // donnee d'entrainement
    Vecteur entree,desiree;
    friend ostream &operator<<(ostream &_os,const Data &_donnee) {
      _donnee.entree.Ecrire(_os); _donnee.desiree.Ecrire(_os); return _os;
    };
    friend istream &operator>>(istream &_is,Data &_donnee) {
      _donnee.entree.Lire(_is); _donnee.desiree.Lire(_is); return _is;
    };
  };

  void    Ecrire(ostream &os=cout);                  // Ecrire les poids
  void    Entrainer(const Table1D<Data> &donnees);   // Entrainer le reseau
  double  EQM(const Vecteur &des);                   // Retourner eqm observee
  double  Gain() {return gain;}                      // Retourner le gain
  void    Gain(double _val) {gain = _val;}           // Definir le gain
  void    InterStat(int _val) {istat = _val;}        // Def. intervalle affich.
  void    Lire(istream &is=cin);                     // Lire les poids
  int     MaxIterations() {return maxiter;}          // Retourner #iterations
  void    MaxIterations(int _val) {maxiter = _val;}  // Definir #iterations
  double  MaxErreur() {return maxerr;}               // Retourner eqm maximum
  void    MaxErreur(double _val) {maxerr = _val;}    // Definir eqm maximum
  double  Momentum() {return momentum;}              // Retourner momentum
  void    Momentum(double _val) {momentum = _val;}   // Definir momentum
  void    PropagerAvant(const Vecteur &entree);      // Propager l'entree
  void    RetroPropager(const Data &donnee);         // Retropropager l'erreur
  Vecteur Sorties();                                 // Retourner les sorties
  double  Tester(const Table1D<Data> &donnees);      // Tester avec tolerance
  double  Tolerance() {return tolerance;}            // Retourner tolerance
  void    Tolerance(double _val) {tolerance = _val;} // Definir tolerance

///////////////////// Fin de l'interface publique //////////////////////////

protected:
  double gain;                              // taux d'apprentissage
  int    istat;                             // intervalle d'affichage
  double maxerr;                            // racine de l'eqm maximum
  int    maxiter;                           // nombre maximum d'iteration
  int    mgraphe;                           // mode graphique
  double momentum;                          // momentum
  int    nentrees;                          // nombre d'entrees
  int    nsorties;                          // nombre de sorties
  double tolerance;                         // tolerance pour les tests

  CProgressDlg fenetre;						// fenetre d'affichage des erreurs

  void InitParam();                         // Initialiser param. entrainement
  void TesterParam();                       // Tester parametres entrainement
};

#endif
