/////////////////////////////////////////////////////////////////////////////
// $Id: Table2D.hpp,v 1.7 1998/01/30 21:02:59 parizeau Exp $
//
//   Auteur:       Marc Parizeau
//   Date:         6/3/96
//   Description:  Patron de classe permettant de grer dynamiquement
//                 un tableau  deux dimensions.
/////////////////////////////////////////////////////////////////////////////

#ifndef TABLE2D_HPP_
#define TABLE2D_HPP_    

#include <iostream.h>
#include <iomanip.h>
#include <limits.h>
#include "Table1D.hpp"

/////////////////////////////////////////////////////////////////////////////

template <class T> class Table2D : public Table1D<T> {

///////////////////// Debut de l'interface publique /////////////////////////

 public:
             Table2D();                       // Construire Table2D 0x0
             Table2D(int i,int j);            // Construire Table2D ixj

  T         &operator()(int i,int j) const;   // Adresser lment

  Table1D<T> Col(int) const;                  // Retourner colonne
  int        Cols() const {return cols;}      // Retourner nombre colonnes
  void       Ecrire(ostream &os=cout) const;  // Ecrire dans flot sortie
  Table1D<T> Ligne(int) const;                // Retourner ligne
  int        Lignes() const {return lignes;}  // Retourner nombre lignes
  void       LignesCols(int i,int j);         // Definir lignes et colonnes
  void       Lire(istream &is=cin);           // Lire dans flot d'entre

  friend ostream &operator<<(ostream &,const Table2D<T> &);
  friend istream &operator>>(istream &,Table2D<T> &);
  
///////////////////// Fin de l'interface publique //////////////////////////

 protected:
  int        lignes;                          // Nombre de lignes 
  int        cols;                            // Nombre de colonnes

};

/////////////////////////////////////////////////////////////////////////////
//  Table2D: Construire Table2D 0x0.
/////////////////////////////////////////////////////////////////////////////
template<class T> 
Table2D<T>::Table2D() {lignes = cols = 0;}

/////////////////////////////////////////////////////////////////////////////
//  Table2D: Construire Table2D ixj.
/////////////////////////////////////////////////////////////////////////////
template<class T> 
Table2D<T>::Table2D(int _i, int _j) : Table1D<T>(_i*_j) {
  lignes = _i;
  cols = _j;
}
   
/////////////////////////////////////////////////////////////////////////
// operator(): Retourner l'lment (i,j).
/////////////////////////////////////////////////////////////////////
template<class T>
inline T &Table2D<T>::operator()(int _i,int _j) const {
#ifndef NDEBUG
  if(_i < 0 || _i >= lignes || _j < 0 || _j >= cols)
    throw ErrBase("Indice invalide!");
#endif
  return pt[_i*cols+_j];
}

/////////////////////////////////////////////////////////////////////////////
// Col: Retourner la colonne spcifie.
/////////////////////////////////////////////////////////////////////////////
template<class T> 
Table1D<T> Table2D<T>::Col(int _j) const {
  Table1D<T> res(lignes);
  for(int i=0;i<lignes;i++) res[i] = (*this)(i,_j);
  return res;
}
   
/////////////////////////////////////////////////////////////////////////
// Ecrire: Ecrire la table dans un flot de sortie.
/////////////////////////////////////////////////////////////////////
template<class T> 
void Table2D<T>::Ecrire(ostream &_os) const {
  int nc=10;
  _os << "(" << lignes << "," << cols << ")" << endl;
  for(int i=0;i<lignes;i++) {
    for(int j=0;j<cols;j++) {
      _os << " " << setw(nc) << (*this)(i,j);
    }
    _os << endl;
  }
}

/////////////////////////////////////////////////////////////////////////////
// Ligne: Retourner la ligne spcifie.
/////////////////////////////////////////////////////////////////////////////
template<class T> 
Table1D<T> Table2D<T>::Ligne(int _i) const {
  Table1D<T> res(cols);
  for(int j=0;j<cols;j++) res[j] = (*this)(_i,j);
  return res;
}
   
/////////////////////////////////////////////////////////////////////////////
//  LignesCols: Dfinir les dimensions de la Table2D.
/////////////////////////////////////////////////////////////////////////////
template<class T> 
void Table2D<T>::LignesCols(int _i, int _j) {
  Dim(_i*_j);
  if(_j > cols) {
    for(int i=lignes-1;i>0;i--) {
      if(i < _i) for(int j=cols-1;j>=0;j--) {
        pt[i*_j+j] = pt[i*cols+j];
      }
    }
  } else if(_j < cols) {
    for(int i=1;i<_i;i++) {
      if(i < lignes) for(int j=0;j<_j;j++) {
        pt[i*_j+j] = pt[i*cols+j];
      }
    }
  }
  lignes = _i;
  cols = _j;
}

/////////////////////////////////////////////////////////////////////////
// Lire: Lire la table dans un flot d'entre.
/////////////////////////////////////////////////////////////////////
template<class T> 
void Table2D<T>::Lire(istream &_flot) {
  char c1=' ',c2=' ',c3=' ';
  int l =-1,c =-1;
  _flot >> c1;
  while(c1 == '#') { // eliminer les lignes de commentaires
    _flot.ignore(INT_MAX,'\n');
    _flot >> c1;
  }
  _flot >> l >> c2 >> c >> c3;
  if(c1 != '(' || c2 != ',' || c3 != ')' || l < 0 || c < 0) 
    throw ErrBase("Format invalide!");
  LignesCols(l,c);
  for(int i=0;i<dim;i++) {
    if(!(_flot >> pt[i])) throw ErrBase("Format invalide!");
  }
}

/////////////////////////////////////////////////////////////////////////
// operator<<: Insrer la table dans le flot de sortie.
/////////////////////////////////////////////////////////////////////////
template<class T>
ostream &operator<<(ostream &_flot,const Table2D<T> &_table) {
    _table.Ecrire(_flot); return _flot;
  }

/////////////////////////////////////////////////////////////////////////
// operator>>: Extraire la table dans le flot d'entre.
/////////////////////////////////////////////////////////////////////////
template<class T>
istream &operator>>(istream &_flot,Table2D<T> &_table) {
    _table.Lire(_flot); return _flot;
  }

#endif
