//////////////////////////////////////////////////////////
// Dfinition du protocole de communication srie 
// pour le 68HC11
//////////////////////////////////////////////////////////
#define LOW_SPEED	30;
#define MED_SPEED	60;
#define HI_SPEED	90;
#define LEFTMOTOR	0;
#define RIGHTMOTOR  1;

//////////////////////////////////////////////////////////
// Dfinition des commandes 
//////////////////////////////////////////////////////////
/*
Commande:	Value:		Param1:			Param2:			Direction:
Stop()		0x01		0				0				PC -> Robot
Rotate()	0x02		angle			vitesse			PC -> Robot
Go()		0x03		distance		vitesse			PC -> Robot
PrintLCD()	0x04		 dterminer	 dterminer	PC -> Robot
Position()	0x05		temps #1		temps #2		Robot -> PC
Distance()	0x06		roue gauche		roue droite		Robot -> PC
*/


//////////////////////////////////////////////////////////
// Dfinition des variables
int dSeq;				// Temps entre la squence de bip
int d;					// Temps entre 2 bip
int c;					// byte lu sur le port
int bccCheck = 0;		// Calcul du bcc du message recu
int msgState = 0;		// Etat du message recu;
char rxMsg[3];			// Le message recu;
char d_gauche;			// Distance parcourue par la roue gauche
char d_droite;			// Distance parcourue par la roue droite



//////////////////////////////////////////////////////////
// Dfinition des fonctions
void ExecuteCmd();				// Lance l'execution de la commande recu
void ProcessData();				// Traite le byte lu sur le port srie
void SendCmd(char,char,char);	// Encodage et envoie d'une commande sur le port 
void Stop();					// Immobilise le robot
void Rotate(int,int);			// Rotation sur place du robot
void Go(int,int);				// Dplacement linaire du robot
void PrintLCD(char*);			// Affichage sur le LCD du robot

/////////////////////////////////////////////////////////////
// Fonction a implmenter
void ReadPort();				// Lecture du port srie
void GetPosition();				// Calcule L1 et L2(temps de rception des bips 1 et 2)
void WritePort(char*);			// Ecriture sur le port srie

///////////////////////////////////////////////////////////
// Dclaration des fonctions

void SendCmd(char cmd, char param1, char param2)
{
	char sendbuf[5];
	char bcc;
	bcc = '$';			// Start byte
	bcc ^= cmd;			// Commande
	bcc ^= param1;		// Paramtre 1
	bcc ^= param2;		// Paramtre 2
	sendbuf[0] = '$';
	sendbuf[1] = cmd;
	sendbuf[2] = param1;
	sendbuf[3] = param2;
	sendbuf[4] = bcc;
	/////////////////////////////
	// Envoie du message
	WritePort(sendbuf);
}

void ProcessData(char c)
{	
	// c: caractre lu sur le port;

	//////////////////////////////////////////////////
	// bcc_check: calcul du bcc du message recu
	// msg:		  buffer de rception du port;
	//
	// msg_state: permet de savoir ou on est rendu dans
	//			  la rception du message
	// Etat			signifie
	// 0			rien recu
	// 1			recu byte de start '$'
	// 2			recu #commande
	// 3			recu param1;
	// 4			recu param2
	// 5			recu bcc;
	switch(msgState)
	{
		case 0: // On a rien recu de valide
			if(c == '$')
			{
				bccCheck = '$';
				msgState++;
			}
			else
				bccCheck = 0;
			break;
		case 1:	// On a recu le start byte
		{
			rxMsg[0] = c;
			bccCheck ^= c;
			msgState++;
		}
			break;
		case 2: // On a recu la commande
		{
			rxMsg[1] = c;
			bccCheck ^= c;
			msgState++;
		}
		case 3: // On a recu le param1
		{
			rxMsg[2] = c;
			bccCheck ^= c;
			msgState++;
		}
		case 4: // On a recu le param2
		{
			if(bccCheck == c)
			{
				// Message Valide
				ExecuteCmd();
			}
			else
			{
				// Message Invalide
				bccCheck = 0;
				msgState = 0;
			}
		}
		default:
			msgState = 0;
			bccCheck = 0;
			break;
	}
}

void ExecuteCmd()
{
	switch(rxMsg[0])
	{
		case 0x01:
			Stop();
			break;
		case 0x02:
			Rotate(rxMsg[1],rxMsg[2]);
			break;
		case 0x03:
			Go(rxMsg[1],rxMsg[2]);
			break;
		case 0x04:
			break;
		case 0x06:
			SendCmd(0x06, d_gauche, d_droite);
			break;
		default:
			break;
	}
}


void Stop()
{
	alloff();
}

void Rotate(int angle, int vitesse)
{
	if(angle == 0) 
		return;
	// L'angle est en degr
	// Ll faut connnaitre la grandeur des roues
	// et la distance entre les roues
	if(angle > 0)
	{
		motor(LEFTMOTOR, vitesse);
		motor(RIGHTMOTOR, -vitesse);
	}
	else
	{
		motor(LEFTMOTOR, -vitesse);
		motor(RIGHTMOTOR, vitesse);
	}
	//while(distance pas atteinte)
	//{}
	Stop();
}

void Go(int distance, int vitesse)
{
	//////////////////////////////
	// On applique les tensions
	motor(LEFTMOTOR, vitesse);
	motor(RIGHTMOTOR, vitesse);
	
	//while(distance pas atteinte)
	//{}
	Stop();
}

void PrintLCD(char *msg)
{
	printf("%s",msg);
}

