//////////////////////////////////////////////////////////////////////////////
// $Id: Liste.hpp,v 1.10 1998/07/15 19:53:18 parizeau Exp $
//
//  Auteur: Marc Parizeau
//
//  Description: Patrons de classe permettant de grer dynamiquement 
//               une liste doublement chane et d'accder  ses
//               lments grace  un ou plusieurs itrateurs. Ce fichier 
//               dfinit les types IterListe et Liste.
//////////////////////////////////////////////////////////////////////////////

#ifndef LISTE_HPP_
#define LISTE_HPP_

#include <iostream.h>
#include "Erreur.hpp"

template<class T> class Liste;

/////////////////////////////////////////////////////////////////////////////

template<class T> class IterListe {

///////////////////// Dbut de l'interface publique /////////////////////////
    
public:
  IterListe(const Liste<T> &liste) : tete(liste.tete), cour(liste.cour) {}

  T         &operator()() const;                 // Retourner courant
  IterListe &operator++();                       // Passer  suivant
  IterListe &operator--();                       // Passer  prcdent
  IterListe &operator=(const IterListe&);        // Affecter itrateur
  int        operator!() const;                  // Tester si courant!=entete
  int        operator==(const IterListe&) const; // Tester quivalence
  int        operator!=(const IterListe&) const; // Tester non-quivalence

  T   &Cour() const;                             // Retourner courant
  int  Dernier() const;                          // Tester si courant=dernier
  int  Entete() const;                           // Tester si courant=entete
  T   &Prec() const;                             // Retourner prcdent
  int  Premier() const;                          // Tester si courant=premier
  int  Queue();                                  // Passer  dernier
  T   &Suiv() const;                             // Retourner suivant
  int  Tete();                                   // Passer  premier
  
///////////////////// Fin de l'interface publique //////////////////////////
  
  class Lien {
  public:

    Lien *suiv;                         // Lien suivant
    Lien *prec;                         // Lien prcdent
    T     element;                      // Elment d'information    
    Lien(Lien *_p=0,Lien *_s=0) {       // Construire un maillon
      prec = _p; suiv = _s;             //  en chanant
      if(prec) prec->suiv = this;       //  avec le prcdent
      if(suiv) suiv->prec = this;       //  et le suivant
    }
    ~Lien() {                           // Dtruire un maillon
      if(prec) prec->suiv = suiv;       //   et refaire la chane
      if(suiv) suiv->prec = prec;       //   de prcdent  suivant
    }
  };

protected:
  IterListe() {tete = cour = 0;}

  Lien *tete;                           // Lien de tte 
  Lien *cour;                           // Lien courant

  friend class Liste<T>;                // ajout pour satisfaire g++

};

/////////////////////////////////////////////////////////////////////////////

template<class T> class Liste : public IterListe<T> {

///////////////////// Dbut de l'interface publique /////////////////////////

public:
  Liste();                                // Construire 
  Liste(const Liste &_l);                 // Construire et copier
  ~Liste() {Vider(); delete tete;}        // Dtruire

  Liste &operator=(const Liste &);        // Affecter liste
  Liste &operator=(const IterListe<T>&);  // Affecter un iterateur
  Liste  operator+(const Liste &) const;  // Fusionner liste
  Liste &operator+=(const Liste &);       // Fusionner/Affecter liste
  T     &operator[](int);                 // Retourner l'lment indic

  int  Chercher(const T&);                // Chercher depuis courant
  int  Chercher(IterListe<T>&,const T&) const; // idem mais avec itrateur
  void Ecrire(ostream &f=cout) const;     // Ecrire dans un flot de sortie
  void Inserer(const T&);                 // Insrer aprs courant
  void Inserer(IterListe<T>&,const T&);   // Insrer aprs courant de itrateur
  void Lire(istream &is=cin);             // Lire dans un flot d'entre
  int  NbElements() const {return nbels;} // Retourner le nombre d'lments  
  void Retirer();                         // Retirer courant
  void Retirer(IterListe<T>&);            // Retirer courant de itrateur
  void Vider();                           // Vider tous les lments
  
  friend ostream &operator<<(ostream &,const Liste<T> &);
  friend istream &operator>>(istream &,Liste<T> &);
  
///////////////////// Fin de l'interface publique //////////////////////////

protected:
  int nbels;                              // nombre d'lments
  T **tpels;                              // table de pointeurs d'lments

};

//////////////////////////////////////////////////////////////////////////////
// operator(): Retourne l'lment courant. Identique  la fonction Cour.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
inline T &IterListe<T>::operator()() const {
  if(cour == tete) throw ErrBase("Pas d'element courant!");
  return cour->element;
}

//////////////////////////////////////////////////////////////////////////////
// operator++: Passer  l'lment suivant.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
inline IterListe<T> &IterListe<T>::operator++() {
  cour = cour->suiv;
  return *this;
}

//////////////////////////////////////////////////////////////////////////////
// operator--: Passer  l'lment prcdent.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
inline IterListe<T> &IterListe<T>::operator--() {
  cour = cour->prec;
  return *this;
}

//////////////////////////////////////////////////////////////////////////////
// operator=: Affecter un itrateur de liste.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
IterListe<T> &IterListe<T>::operator=(const IterListe<T> &_iter) {
  if(tete != _iter.tete) throw ErrBase("Iterateurs invalides!");
  cour = _iter.cour;
  return *this;
}

//////////////////////////////////////////////////////////////////////////////
// operator!: Retourner entier non nul si courant existe, et 0 autrement.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
inline int IterListe<T>::operator!() const {return cour != tete;}

//////////////////////////////////////////////////////////////////////////////
// operator==: Retourner entier non nul si itrateur quivalent, 0 autrement.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
int IterListe<T>::operator==(const IterListe<T> &_iter) const {
  return tete == _iter.tete && cour == _iter.cour;
}

//////////////////////////////////////////////////////////////////////////////
// operator!=: Retourner entier non nul si itrateur non-quiv., 0 autrement.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
int IterListe<T>::operator!=(const IterListe<T> &_iter) const {
  return tete != _iter.tete || cour != _iter.cour;
}

//////////////////////////////////////////////////////////////////////////////
// Cour: Retourner l'lment courant.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
inline T &IterListe<T>::Cour() const {
  if(cour == tete) throw ErrBase("Pas d'element courant!");
  return cour->element;
}

//////////////////////////////////////////////////////////////////////////////
// Dernier: Retourner entier non nul si courant=dernier, 0 autrement.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
inline int IterListe<T>::Dernier() const {
  return (cour != tete && cour == tete->prec);
}

//////////////////////////////////////////////////////////////////////////////
// Entete: Retourner entier non nul si courant=entete, 0 autrement.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
inline int IterListe<T>::Entete() const {
  return cour == tete;
}

//////////////////////////////////////////////////////////////////////////////
// Prec: Retourner l'lment prcdent l'lment courant.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
inline T &IterListe<T>::Prec() const {
  if(cour->prec == tete) throw ErrBase("Pas d'element precedent!");
  return cour->prec->element;
}

//////////////////////////////////////////////////////////////////////////////
// Premier: Retourner entier non nul si courant=premier, 0 autrement.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
inline int IterListe<T>::Premier() const {
  return (cour != tete && cour == tete->suiv);
}

//////////////////////////////////////////////////////////////////////////////
// Queue: Passer au dernier lment de la liste (cour=dernier).
//////////////////////////////////////////////////////////////////////////////
template<class T> 
inline int IterListe<T>::Queue() {
  cour = tete->prec; 
  return cour != tete;
}

//////////////////////////////////////////////////////////////////////////////
// Suiv: Retourner l'lment suivant l'lment courant.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
inline T &IterListe<T>::Suiv() const {
  if(cour->suiv == tete) throw ErrBase("Pas d'element suivant!");
  return cour->suiv->element;
}

//////////////////////////////////////////////////////////////////////////////
// Tete: Passer au premier lment de la liste (cour=premier).
//////////////////////////////////////////////////////////////////////////////
template<class T> 
inline int IterListe<T>::Tete() {
  cour = tete->suiv; 
  return cour != tete;
}

//////////////////////////////////////////////////////////////////////////////
// Liste: Construire une liste.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
Liste<T>::Liste() : IterListe<T>(*this) {
  tete = cour = new Lien;            // allouer lment d'entete
  cour->suiv = cour->prec = tete;    // chainer la liste
  tpels = 0; nbels = 0;
}

//////////////////////////////////////////////////////////////////////////////
// Liste: Construire et copier une liste. Attention, l'itrateur de
//        la nouvelle liste ne pointe pas ncessairement sur l'lment
//        correspond de l'ancienne!
//////////////////////////////////////////////////////////////////////////////
template<class T> 
Liste<T>::Liste(const Liste<T> &_l) : IterListe<T>(*this) {
  tete = cour = new Lien;            // allouer lment d'entete
  cour->suiv = cour->prec = tete;    // chainer la liste
  tpels = 0; nbels = 0;
  operator=(_l);
}

//////////////////////////////////////////////////////////////////////////////
// operator=: Affecter une liste.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
Liste<T> &Liste<T>::operator=(const Liste<T> &_liste) {
  if(&_liste == this) return *this;
  Vider();
  IterListe<T> tmp(_liste);
  for(tmp.Tete();!tmp;++tmp) Inserer(tmp());
  return *this;
}

//////////////////////////////////////////////////////////////////////////////
// operator=: Affecter un iterateur.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
Liste<T> &Liste<T>::operator=(const IterListe<T> &_iter) {
  IterListe<T>::operator=(_iter);
  return *this;
}

//////////////////////////////////////////////////////////////////////////////
// operator+: Fusionner avec une liste.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
Liste<T> Liste<T>::operator+(const Liste<T> &_liste) const {
  Liste<T> res = (*this);
  res.Queue();
  IterListe<T> tmp(_liste);
  for(tmp.Tete();!tmp;++tmp) res.Inserer(tmp());
  return res;
}

//////////////////////////////////////////////////////////////////////////////
// operator+=: Fusionner/Affecter avec une liste.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
Liste<T> &Liste<T>::operator+=(const Liste<T> &_liste) {
  Queue();
  IterListe<T> tmp(_liste);
  for(tmp.Tete();!tmp;++tmp) Inserer(tmp());
  return *this;
}

//////////////////////////////////////////////////////////////////////////////
// operator[]: Adresser l'lment indic.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
inline T &Liste<T>::operator[](int _n) {
  if(!tpels) {
    tpels = new T*[nbels];             // contruire la table des pointeurs
    if(!tpels) throw ErrBase("Erreur d'allocation de memoire!");
    IterListe<T> tmp(*this);
    int i=0;
    for(tmp.Tete();!tmp;++tmp) tpels[i++] = &tmp();
  }
  if(_n < 0 || _n >= nbels) throw ErrBase("Indice invalide!");
  return *tpels[_n];
}

//////////////////////////////////////////////////////////////////////////////
// Chercher: Chercher  partir de l'lment courant.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
int Liste<T>::Chercher(const T &_elem) {
  for(;cour!=tete;cour=cour->suiv) if(_elem == cour->element) break;
  return cour != tete;
}

//////////////////////////////////////////////////////////////////////////////
// Chercher: Chercher  partir de l'lment point par l'itrateur.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
int Liste<T>::Chercher(IterListe<T> &_iter,const T &_elem) const {
  if(tete != _iter.tete) throw ErrBase("Iterateur incompatible!");
  for(;!_iter;++_iter) if(_elem == _iter()) break;
  return !_iter;
}

//////////////////////////////////////////////////////////////////////////////
// Ecrire: Ecrire la liste dans le flot de sortie (mme format que Table1D).
//////////////////////////////////////////////////////////////////////////////
template<class T> 
void Liste<T>::Ecrire(ostream &_flot) const {
  _flot << "(" << NbElements() << ")" << endl;
  IterListe<T> iter(*this);
  for(iter.Tete();!iter;++iter) _flot << iter() << endl;
  _flot << endl;
}

//////////////////////////////////////////////////////////////////////////////
// Inserer: Insrer un lment apres l'lment courant.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
void Liste<T>::Inserer(const T &_elem) {
  if(tpels) {delete tpels; tpels = 0;}
  cour = new Lien(cour,cour->suiv);
  if(cour == 0) throw ErrBase("Erreur d'allocation de memoire!");
  cour->element = _elem;
  nbels++;
}

//////////////////////////////////////////////////////////////////////////////
// Inserer: Insrer un lment aprs l'lment pointe par l'itrateur.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
void Liste<T>::Inserer(IterListe<T> &_iter,const T &_elem) {
  if(tete != _iter.tete) throw ErrBase("Iterateur incompatible!");
  if(tpels) {delete tpels; tpels = 0;}
  _iter.cour = new Lien(_iter.cour,_iter.cour->suiv);
  if(_iter.cour == 0) throw ErrBase("Erreur d'allocation de memoire!");
  _iter.cour->element = _elem;
  nbels++;
}

//////////////////////////////////////////////////////////////////////////////
// Lire: Lire une liste dans un flot d'entre (mme format que Table1D).
//////////////////////////////////////////////////////////////////////////////
template<class T> 
void Liste<T>::Lire(istream &_flot) {
  int n; char c1=' ',c2=' ';
  _flot >> c1 >> n >> c2;
  if(c1 != '(' || c2 != ')' || n < 0) throw ErrBase("Format invalide!");
  Vider(); T tmp;
  for(int i=0;i<n;i++) {
    if(!(_flot >> tmp)) throw ErrBase("Format invalide!");
    Inserer(tmp);
  }
}

//////////////////////////////////////////////////////////////////////////////
// Retirer: Retirer l'lment courant.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
void Liste<T>::Retirer() {
  if(tpels) {delete tpels; tpels = 0;}
  if(cour == tete) throw ErrBase("Pas d'element courant!");
  cour = cour->prec;
  Lien *tmp = cour->suiv;                   // tape ncessaire pour -fast
  delete tmp;
  nbels--;
}

//////////////////////////////////////////////////////////////////////////////
// Retirer: Retirer l'lment courant.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
void Liste<T>::Retirer(IterListe<T> &_iter) {
  if(tete != _iter.tete) throw ErrBase("Iterateur incompatible!");
  if(tpels) {delete tpels; tpels = 0;}
  if(_iter.cour == tete) throw ErrBase("Pas d'element courant!");
  if(cour == _iter.cour) cour = _iter.cour->prec;
  _iter.cour = _iter.cour->prec;
  Lien *tmp = _iter.cour->suiv;             // tape ncessaire pour -fast
  delete tmp;
  nbels--;
}

//////////////////////////////////////////////////////////////////////////////
// Vider: Vider la liste en retirant tous ses lments.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
void Liste<T>::Vider() {
  if(tpels) {delete tpels; tpels = 0;}
  cour = tete;
  while(cour->suiv != tete) {
    Lien *tmp = cour->suiv;                 // tape ncessaire pour -fast
    delete tmp;
  }
  nbels = 0;
}
  
/////////////////////////////////////////////////////////////////////////
// operator<<: Insrer la liste dans le flot de sortie.
/////////////////////////////////////////////////////////////////////////
template<class T>
ostream &operator<<(ostream &_flot,const Liste<T> &_liste) {
  _liste.Ecrire(_flot); return _flot;
}

/////////////////////////////////////////////////////////////////////////
// operator>>: Extraire la liste dans le flot d'entre.
/////////////////////////////////////////////////////////////////////////
template<class T>
istream &operator>>(istream &_flot,Liste<T> &_liste) {
  _liste.Lire(_flot); return _flot;
}

#endif
