//////////////////////////////////////////////////////////////////////////////
// $Id: MonceauBin.hpp,v 1.4 1998/01/30 21:02:59 parizeau Exp $
//
//  Auteur: Marc Parizeau
//  Date: 25/11/96
//  Description: Patron de classe permettant de gerer dynamiquement 
//               un monceau binaire.
//////////////////////////////////////////////////////////////////////////////

#ifndef MONCEAU_HPP_
#define MONCEAU_HPP_

#include <iostream.h>
#include "Table1D.hpp"
#include "Erreur.hpp"

/////////////////////////////////////////////////////////////////////////////

template<class T> class MonceauBin : protected Table1D<T> {
    
///////////////////// Debut de l'interface publique /////////////////////////
    
public:
  MonceauBin(int _dim=100) : Table1D<T>(_dim+1) {n = 0; ordre = 1;}
    
  T   &ChercherMin();                   // Retourner element minimum
  void Entasser(const T &);             // Inserer element (sans ordre)
  void Inserer(const T &);              // Inserer element (avec ordre)
  int  NbElements() const {return n;}   // Retourner nombre d'elements
  void Ordonner();                      // Retablir propriete d'ordre
  T    RetirerMin();                    // Retirer minimum
  void RetirerMin(T &);                 // Retirer minimum
  void Vider() {n = 0;}                 // Vider monceau

///////////////////// Fin de l'interface publique //////////////////////////

protected:
  int n;                                // nombre d'elements
  int ordre;                            // indique si les elements sont ordonnes

private:
  MonceauBin(MonceauBin &) {}           // pour interdire les recopies
  void PercolerBas(int i);              // Pecoler vers le bas
    
};

//////////////////////////////////////////////////////////////////////////////
// ChercherMin: Retourner l'element minimum.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
T &MonceauBin<T>::ChercherMin() {
  if(n == 0) throw ErrBase("Pas d'element minimum!");
  if(!ordre) Ordonner();
  return pt[1];
}

//////////////////////////////////////////////////////////////////////////////
// Entasser: Inserer un element sans maintenir la propriete d'ordre.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
void MonceauBin<T>::Entasser(const T &_element) {
  if(n+1 == Dim()) Dim(2*Dim());       // aggrandir le monceau!
  pt[++n] = _element;
  if(_element < pt[n/2]) ordre = 0;
}

//////////////////////////////////////////////////////////////////////////////
// Inserer: Inserer un element en maintenant la propriete d'ordre.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
void MonceauBin<T>::Inserer(const T &_element) {
  if(!ordre) Entasser(_element);
  else {
    if(n+1 == Dim()) Dim(2*Dim());       // aggrandir le monceau!
    int i = ++n;
    while(i > 1 && pt[i/2] > _element) { // percoler vers le haut
      pt[i] = pt[i/2];
      i /= 2;
    }
    pt[i] = _element;
  }
}

//////////////////////////////////////////////////////////////////////////////
// Ordonner: Retablir la propriete d'ordre du monceau.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
void MonceauBin<T>::Ordonner() {
  for(int i=n/2; i>0; i--) PercolerBas(i);
  ordre = 1;
}

//////////////////////////////////////////////////////////////////////////////
// PercolerBas: Percoler vers le bas la position i.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
void MonceauBin<T>::PercolerBas(int _i) {
  int ppfils;                 // plus petit fils
  T tmp = pt[_i];
  for(;_i*2 <= n;_i=ppfils) { // choisir le ppfils et percoler vers le bas
    ppfils = _i*2;
    if(ppfils != n && pt[ppfils+1] < pt[ppfils]) ppfils++;
    if(pt[ppfils] < tmp) pt[_i] = pt[ppfils];
    else break;
  }
  pt[_i] = tmp;
}

//////////////////////////////////////////////////////////////////////////////
// RetirerMin: Retirer et retourner l'element minimum.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
T MonceauBin<T>::RetirerMin() {
  T res;
  RetirerMin(res);
  return res;
}

//////////////////////////////////////////////////////////////////////////////
// RetirerMin: Retirer et retourner l'element minimum.
//////////////////////////////////////////////////////////////////////////////
template<class T> 
void MonceauBin<T>::RetirerMin(T &_res) {
  _res = ChercherMin();
  pt[1] = pt[n--];
  PercolerBas(1);
}

#endif
