/////////////////////////////////////////////////////////////////////////////
// $Id: Random.cpp,v 1.2 1997/08/20 02:56:35 parizeau Exp $
/////////////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <math.h>
#include <time.h>
#include "Random.hpp"

Random _gnpa;      // Generateur de Nombres Pseudo-Aleatoire (variable globale)

/////////////////////////////////////////////////////////////////////////////
// Random: Construire le generateur.
/////////////////////////////////////////////////////////////////////////////
Random::Random(long _init) {
  if(_init <= 0) idum = time(0);
  else idum = _init;
  for(int j=NTAB+7;j>=0;j--) {
    long k = idum/IQ;
    idum = IA*(idum-k*IQ)-IR*k;
    if(idum < 0) idum += IM;
    if(j < NTAB) iv[j] = idum;
  }
  iy = iv[0];
  iset = 0;
}

/////////////////////////////////////////////////////////////////////////////
// Entier: Generer un entier dans l'itervalle [a,b].
/////////////////////////////////////////////////////////////////////////////
int Random::Entier(int _a, int _b) {
  int e = (int) (_a + (_b - _a + 1) * Unif());
  if(e > _b) e = _b;
  return e;
}

/////////////////////////////////////////////////////////////////////////////
// Unif: Retourne une valeur distribuee uniformement dans l'intervalle [0,1].
/////////////////////////////////////////////////////////////////////////////
double Random::Unif() {
  double temp;

  long k = idum/IQ;
  idum = IA*(idum-k*IQ)-IR*k;
  if(idum < 0) idum += IM;
  int j = (int) iy/NDIV;
  iy = iv[j];
  iv[j] = idum;
  if((temp=AM*iy) > RNMX) return RNMX;
  else return temp;
}

/////////////////////////////////////////////////////////////////////////////
// Gauss: Retourne une valeur distribuee selon une normalle N(_mu,_sigma).
/////////////////////////////////////////////////////////////////////////////
double Random::Gauss(double _mu,double _sigma) {
  double fac,rsq,v1,v2;

  if(iset == 0) {
    do {
      v1 = 2.0*Unif()-1.0;
      v2 = 2.0*Unif()-1.0;
      rsq = v1*v1+v2*v2;
    } while(rsq >= 1.0 || rsq == 0);
    fac = sqrt(-2.0*log(rsq)/rsq);
    gset = v1*fac;
    iset = 1;
    return v2*fac*_sigma+_mu;
  } else {
    iset = 0;
    return gset*_sigma+_mu;
  }
}

