//////////////////////////////////////////////////////////////////////////////
// $Id: Random.hpp,v 1.2 1997/08/20 02:57:58 parizeau Exp $
//
//  Auteur:       Marc Parizeau
//  Creation:     17 novemmbre 1994
//  Modification: 8 mars 1996
//
//  Description: 
//
// Cette classe permet d'initialiser et de gerer un generateur de
// nombre aleatoire de distribution uniforme ou gaussienne. Les methodes
// sont tirees du livre numerical recipes en C (rand1).
//
//////////////////////////////////////////////////////////////////////////////

#ifndef RANDOM_HPP_
#define RANDOM_HPP_

#define IA 16807
#define IM 2147483647
#define AM (1.0/IM)
#define IQ 127773
#define IR 2836
#define NTAB 32
#define NDIV (1+(IM-1)/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)

/////////////////////////////////////////////////////////////////////////////

class Random {

///////////////////// Debut de l'interface publique /////////////////////////
 
public:
  Random(long init=0);                   // Initialiser (defaut: horloge)

  int    Entier(int a, int b);           // entier dans [a,b]
  double Unif();                         // distribution uniforme dans [0,1]
  double Gauss(double m=0,double s=1);   // distribution gaussienne N(m,s)

///////////////////// Fin de l'interface publique //////////////////////////
 
 private:

  long iy;
  long iv[NTAB];
  long idum;

  int iset;
  double gset;

};

extern Random _gnpa;   // Generateur de Nombres Pseudo-Aleatoires
                       // Attention, cette variable est globale!

#endif



