/////////////////////////////////////////////////////////////////////////////
// $Id: Table1D.hpp,v 1.11 1998/04/15 21:22:36 parizeau Exp $
//
//   Auteur:       Marc Parizeau
//   Description:  Patron de classe permettant de grer dynamiquement
//                 un tableau  une dimension. Ce fichier dfinit les
//                 types Table1D et TableInt.
/////////////////////////////////////////////////////////////////////////////

#ifndef TABLE1D_HPP_
#define TABLE1D_HPP_    

#include <iostream.h> 
#include <iomanip.h>
#include <stdlib.h>
#include <limits.h>
#include "Random.hpp"
#include "Erreur.hpp"

/////////////////////////////////////////////////////////////////////////////

template <class T> class Table1D {

///////////////////// Dbut de l'interface publique /////////////////////////
  
 public:
  Table1D(int=0);                        // Construire
  Table1D(int,const T&);                 // Construire et initialiser
  Table1D(const Table1D&);               // Construire et copier
  ~Table1D();                            // Dtruire

  Table1D &operator=(const Table1D&);    // Affecter une table
  T   &operator[](int i) const;          // adresser lment
  int  operator==(const Table1D&) const; // Tester quivalence
  int  operator!=(const Table1D&) const; // Tester non-quivalence

  int  Chercher(const T&) const;         // Faire recherche binaire
  void Dim(int);                         // Dfinir dimension
  int  Dim() const {return dim;}         // Retourner dimension effective
  int  Dir() const {return dir;}         // Retourner dimension relle
  void Ecrire(ostream &os=cout) const;   // Ecrire dans flot de sortie
  void Initialiser(const T&);            // Initialiser valeur
  void Lire(istream &os=cin);            // Lire dans flot d'entre
  void Permuter(Random &rd=_gnpa);       // Permuter 
  void Trier(int=10);                    // Trier avec QuickSort

  friend ostream &operator<<(ostream &,const Table1D<T> &);
  friend istream &operator>>(istream &,Table1D<T> &);

///////////////////// Fin de l'interface publique //////////////////////////

 protected:
  int dir;                              // dimension reelle
  int dim;                              // Dimension effective
  T   *pt;                              // tableau des lments

 private:
  void  ISort(int);                     // trier par insertion
  T    &Median3(int,int);               // trouver mdiane de 3
  void  QSort(int,int,int);             // quicksort
  void  Swap(T &a,T &b) {T tmp = a; a = b; b = tmp;}

  void  operator=(int) {}  // pour viter les conversions int->Table1D

};

/////////////////////////////////////////////////////////////////////////////

typedef Table1D<int> TableInt;

/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// Table1D: Construire.
/////////////////////////////////////////////////////////////////////////////
template<class T> 
Table1D<T>::Table1D(int _dim) {
  pt = 0;
  Dim(_dim);  
}

/////////////////////////////////////////////////////////////////////////////
// Table1D: Construire et initialiser.
/////////////////////////////////////////////////////////////////////////////
template<class T> 
Table1D<T>::Table1D(int _dim,const T &_val) {
  pt = 0;
  Dim(_dim);  
  Initialiser(_val);
}

/////////////////////////////////////////////////////////////////////////////
// Table1D: Construire et copier
/////////////////////////////////////////////////////////////////////////////
template <class T> 
Table1D<T>::Table1D(const Table1D<T> &_table) {
   pt = 0;
   (*this) = _table;
}

/////////////////////////////////////////////////////////////////////////////
// ~Table1D: Destructeur.
/////////////////////////////////////////////////////////////////////////////
template<class T> 
Table1D<T>::~Table1D() {
  if(pt) delete [] pt;
  pt = 0;
}

/////////////////////////////////////////////////////////////////////////////
// operator=: Affecter un Table1D.
/////////////////////////////////////////////////////////////////////////////
template<class T>  
Table1D<T> &Table1D<T>::operator=(const Table1D<T> &_table) {
  if(&_table == this) return *this;
  Dim(_table.dim);
  for(int i=0;i<dim;i++) pt[i] = _table[i];
  return *this;
}

/////////////////////////////////////////////////////////////////////////////
// operator[]: Retourner l'lment indic.
/////////////////////////////////////////////////////////////////////////////
template<class T> 
inline T &Table1D<T>::operator[](int _i) const {
  if(_i < 0 || _i >= dim) throw ErrBase("Indice invalide!");
  return pt[_i];
}

/////////////////////////////////////////////////////////////////////////////
// operator==: Tester l'quivalence avec un Table1D.
/////////////////////////////////////////////////////////////////////////////
template<class T> 
int Table1D<T>::operator==(const Table1D<T> &_table) const {
  if(dim != _table.dim) return 0;
  for(int i=0;i<dim;i++) if(pt[i] != _table[i]) return 0;
  return 1;
}

/////////////////////////////////////////////////////////////////////////////
// operateur!=: Tester la non-quivalence avec un Table1D.
/////////////////////////////////////////////////////////////////////////////
template<class T> 
int Table1D<T>::operator!=(const Table1D<T> &_table) const {
  if(dim != _table.dim) return 1;
  for(int i=0;i<dim;i++) if(pt[i] != _table[i]) return 1;
  return 0;
}

///////////////////////////////////////////////////////////////////////////
// Chercher: Faire une recherche binaire (la table doit tre trie) et 
//           retourner l'indice de l'lment dans la table  
//           (-1 si lment non trouv).
///////////////////////////////////////////////////////////////////////////
template<class T>
int Table1D<T>::Chercher(const T &_element) const {
  int deb = 0, fin = Dim()-1;
  int mil;
  while(deb <= fin) {
    mil = (deb+fin)/2;
    if((*this)[mil] < _element) deb = mil+1;
    else if((*this)[mil] > _element) fin = mil-1;
    else return mil;
  }
  return -1;
}

/////////////////////////////////////////////////////////////////////////////
// Dim: Definir la dimension du Table1D.
/////////////////////////////////////////////////////////////////////////////
template<class T> 
void Table1D<T>::Dim(int _dim) {
  if(_dim < 0) throw ErrBase("Dimension invalide!");
  if((!pt && _dim > 0) || (pt && _dim > dir)) {
    T *tmp=pt;
    pt = new T[_dim];
    if(!pt) throw ErrBase("Erreur d'allocation de memoire!");
    if(tmp) {
      for(int i=0;i<dim;i++) pt[i] = tmp[i];
      delete [] tmp;
    }
    dir = _dim;
  } else if(pt && _dim == 0) {
    delete [] pt;
    dir = 0; pt = 0;
  }
  dim = _dim;
}
  
/////////////////////////////////////////////////////////////////////////////
// Ecrire: Ecrire une Table1D.
/////////////////////////////////////////////////////////////////////////////
template<class T> 
void Table1D<T>::Ecrire(ostream &_flot) const {
  _flot << "(" << dim << ")" << endl;
  for(int i=0;i<dim;i++) _flot << pt[i] << endl;
}

/////////////////////////////////////////////////////////////////////////////
// Initialiser: Initialiser la table avec une valeur.
/////////////////////////////////////////////////////////////////////////////
template<class T> 
void Table1D<T>::Initialiser(const T &_val) {
  for(int i=0;i<dim;i++) pt[i] = _val;
}
   
/////////////////////////////////////////////////////////////////////////
// Lire: Lire une table.
/////////////////////////////////////////////////////////////////////////
template<class T> 
void Table1D<T>::Lire(istream &_flot) {
  char c1=' ',c2=' ';
  int n = -1;
  _flot >> c1;
  while(c1 == '#') { // eliminer les lignes de commentaires
    _flot.ignore(INT_MAX,'\n');
    _flot >> c1;
  }
  _flot >> n >> c2;
  if(c1 != '(' || c2 != ')' || n < 0) throw ErrBase("Format invalide!");
  Dim(n);
  for(int i=0;i<n;i++) {
    if(!(_flot >> pt[i])) throw ErrBase("Format invalide!");
  }
}

/////////////////////////////////////////////////////////////////////////
// Permuter: Permuter alatoirement la table.
/////////////////////////////////////////////////////////////////////////
template<class T>
void Table1D<T>::Permuter(Random &_rd) {
  int j;
  for(int i=1;i<Dim();i++) {
    j = _rd.Entier(0,i);
    Swap(pt[i],pt[j]);
  }
}

/////////////////////////////////////////////////////////////////////////
// Trier: Trier la table avec l'algorithme de tri rapide (quicksort).
/////////////////////////////////////////////////////////////////////////
template<class T>
void Table1D<T>::Trier(int _cut) {
  if(_cut > Dim()) _cut = Dim();
  QSort(0,Dim()-1,_cut);
  ISort(_cut);
}

/////////////////////////////////////////////////////////////////////////
// ISort: Tri par insertion adapt pour le quicksort avec seuil=cut.
/////////////////////////////////////////////////////////////////////////
template<class T>
void Table1D<T>::ISort(int cut) {
  int i,p=0;
  for(i=1;i<cut;i++) {             // chercher le min pour la sentinelle
    if((*this)[i] < (*this)[p]) p = i;
  }
  Swap((*this)[0],(*this)[p]);
  T tmp;
  for(p=1;p<Dim();p++) {
    tmp = (*this)[p];
    for(i=p;tmp<(*this)[i-1];i--) (*this)[i] = (*this)[i-1];
    (*this)[i] = tmp;
  }
}

/////////////////////////////////////////////////////////////////////////
// Median3: Choisir la mdiane des lments gauche, centre et droit.
/////////////////////////////////////////////////////////////////////////
template<class T>
T &Table1D<T>::Median3(int left,int right) {
  int center = (left + right)/2;
  if((*this)[left] > (*this)[center]) Swap((*this)[left],(*this)[center]);
  if((*this)[left] > (*this)[right]) Swap((*this)[left],(*this)[right]);
  if((*this)[center] > (*this)[right]) Swap((*this)[center],(*this)[right]);
  Swap((*this)[center],(*this)[right-1]);
  return (*this)[right-1];
}

/////////////////////////////////////////////////////////////////////////
// QSort: fonction recursive du Quicksort. (voir M.A. Weiss, page 273)
/////////////////////////////////////////////////////////////////////////
template<class T>
void Table1D<T>::QSort(int left,int right,int cut) {
  if(left + cut <= right) {
    T pivot = Median3(left,right);
    int i = left, j = right-1;
    for(;;) {
      while((*this)[++i] < pivot);
      while((*this)[--j] > pivot);
      if(i < j) Swap((*this)[i],(*this)[j]);
      else break;
    }
    Swap((*this)[i],(*this)[right-1]);
    QSort(left,i-1,cut);
    QSort(i+1,right,cut);
  }
}

/////////////////////////////////////////////////////////////////////////
// operator<<: Inserer la table dans le flot de sortie.
/////////////////////////////////////////////////////////////////////////
template<class T>
ostream &operator<<(ostream &_flot,const Table1D<T> &_table) {
    _table.Ecrire(_flot); return _flot;
  }

/////////////////////////////////////////////////////////////////////////
// operator>>: Extraire la table dans le flot d'entre.
/////////////////////////////////////////////////////////////////////////
template<class T>
istream &operator>>(istream &_flot,Table1D<T> &_table) {
    _table.Lire(_flot); return _flot;
  }

#endif
